module CalcLike
use cosmonet
 use CMB_Cls
 use cmbtypes
 use cmbdata
 use mpk
 use Random
 use settings
 use ParamDef
 use snovae
 use WeakLen
 use Lya
 implicit none

 logical :: Use_HST = .true.
 logical :: Use_Age_Tophat_Prior = .true.
 logical :: Use_CMB = .true.
 logical :: Use_BBN = .false.
 logical :: Use_Clusters = .false.
 
 integer :: H0_min = 40, H0_max = 100
 real :: Omk_min = -0.3, Omk_max = 0.3
 real :: Use_min_zre = 0
 integer :: Age_min = 10, Age_max = 20
 real :: Temperature  = 1

contains
  
  function GetLogPrior(CMB, Info) !Get -Ln(Prior)
    real GetLogPrior
    real Age
    Type (CMBParams) CMB
    Type(ParamSetInfo) Info

    GetLogPrior = logZero
 
    if (CMB%H0 < H0_min .or. CMB%H0 > H0_max) return
    if (CMB%omk < Omk_min .or. CMB%omk > Omk_max .or. CMB%Omv < 0) return
    if (CMB%zre < Use_min_zre) return

    Age = GetAge(CMB, Info)
     !This sets up parameters in CAMB, so do not delete unless you are careful!
 
    if (Use_Age_Tophat_Prior .and. (Age < Age_min .or. Age > Age_max) .or. Age < 0) return
    GetLogPrior = 0
 
  end function GetLogPrior

function COSMONET_GetLogLike(Params, time)
	use cosmonet
	logical time, use_cosmonet
	 Type (CMBParams) CMB
	Type (ParamSet) Params
    real COSMONET_GetLogLike
    real dum(1,1), regression, net_params(1:2*num_params_used)
        integer error, i
        real acl(lmax,num_cls)
	real test_acl(lmax, num_cls)                   
                                                               
        if (time) then
                open(35, FILE = NETPARAMS)
		!write(*,*) num_params_used
                read(35,*) net_params
                
		FScales%PMin = Scales%PMin
                FScales%PMax = Scales%PMax
                do i=1, num_params_used
                        FScales%PMin(params_used(i)) = net_params(i)
                        FScales%PMax(params_used(i)) = net_params(i+num_params_used)
                end do
                                                                                                              
        end if

    use_cosmonet = .true.
    if (any(Params%P > FScales%PMax) .or. any(Params%P < FScales%PMin)) then
        use_cosmonet = .false.
    else
        use_cosmonet = .true.
    end if


if (use_cosmonet == .false. .or. user_cosmonet == .false.) then
    if (any(Params%P > Scales%PMax) .or. any(Params%P < Scales%PMin)) then
       COSMONET_GetLogLike = logZero
        return
    end if
                                                                                  
    call ParamsToCMBParams(Params%P,CMB)
                                                                                  
    COSMONET_GetLogLike = GetLogLikePost(CMB, Params%Info,dum,.false.)
else
	
!!!Set up parameters each time!!!
                                                                                               
	do i = 1, num_params_used
        	COSMONETP(i) = Params%P(params_used(i))
	end do

    if (any(Params%P > Scales%PMax) .or. any(Params%P < Scales%PMin)) then
       COSMONET_GetLogLike = logZero
        return
    end if

    if (time) then
        write(*,*) "INITIALISE FIRST TIME"
        call COSMONET_GetCls(CMB, Params%Info, acl, error, time)
        COSMONET_GetLogLike = GetLogLike(Params)
        return
    end if
                                                                                               
	call ParamsToCMBParams(Params%P,CMB)
	COSMONET_GetLogLike = COSMONET_GetLogLikePost(CMB, Params%Info,dum,.false., time, use_cosmonet)

end if                                                                        
   end function COSMONET_GetLogLike





  function GetLogLike(Params) !Get -Ln(Likelihood)
    type(ParamSet)  Params 
    Type (CMBParams) CMB
    real GetLogLike
    real dum(1,1)
 
    if (any(Params%P > Scales%PMax) .or. any(Params%P < Scales%PMin)) then
       GetLogLike = logZero
        return
    end if

    call ParamsToCMBParams(Params%P,CMB)

    GetLogLike = GetLogLikePost(CMB, Params%Info,dum,.false.)
   
   end function GetLogLike

 function COSMONET_GetLogLikePost(CMB, Info, inCls, HasCls, time, use_cosmonet)
   real COSMONET_GetLogLikePost
    Type (CMBParams) CMB
    type(ParamSet)  Params
    Type(ParamSetInfo) Info
    real, intent(in):: inCls(:,:)
    logical, intent(in) :: HasCls
        logical time, use_cosmonet
    real acl(lmax,num_cls)
    integer error
                                                                                                                             
    COSMONET_GetLogLikePost  = GetLogPrior(CMB, Info)
    if ( COSMONET_GetLogLikePost >= logZero) then
       COSMONET_GetLogLikePost = logZero
                                                                                                                             
else
                                                                                                                             
       if (Use_HST) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + (CMB%H0 - 72)**2/(2*8**2)  !HST
       if (Use_BBN) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + (CMB%ombh2 - 0.022)**2/(2*0.002**2)
          !I'm using increased error bars here
                                                                                                                             
       if (Use_CMB .or. Use_LSS) then
          if (HasCls) then
           acl = inCls
         error =0
          else if (Use_LSS) then
           	call GetCls(CMB, Info, acl, error)
          else
	    	call COSMONET_GetCls(CMB, Info, acl, error, time)
	 end if
         if (error /= 0) then
          COSMONET_GetLogLikePost = logZero
  else
		COSMONET_GetLogLikePost = CMBLnLike(acl, use_cosmonet) + COSMONET_GetLogLikePost
        if (Use_mpk) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + LSSLnLike(CMB, Info%theory)
          if (Use_WeakLen) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + WeakLenLnLike(CMB, Info%theory)
          if (Use_Lya) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost +  LSS_Lyalike(CMB, Info%Theory)
          if ( COSMONET_GetLogLikePost >= logZero) then
            COSMONET_GetLogLikePost = logZero
          end if
         end if
         if (Use_SN .and. COSMONET_GetLogLikePost /= logZero ) then
           if (Info%Theory%SN_loglike /= 0) then
                COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + Info%Theory%SN_loglike
            else
             COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + SN_LnLike(CMB)
            end if
               !Assume computed only every time hard parameters change
                                                                                                                             
                                                                                                                             
         end if
                                                                                                                             
       else
         if (Use_SN) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + SN_LnLike(CMB)
       end if
                                                                                                                             
      if (Use_Clusters .and. COSMONET_GetLogLikePost /= LogZero) then
          COSMONET_GetLogLikePost = COSMONET_GetLogLikePost + &
                 (Info%Theory%Sigma_8-0.9)**2/(2*0.05**2)
          stop 'Write your cluster prior in calclike.f90 first!'
      end if
                                                                                                                             
     if (COSMONET_GetLogLikePost /= LogZero) COSMONET_GetLogLikePost = COSMONET_GetLogLikePost/Temperature
    end if
                                                                                                                             
  end function COSMONET_GetLogLikePost


    
  function GetLogLikePost(CMB, Info, inCls, HasCls) 
    real GetLogLikePost
    Type (CMBParams) CMB
    Type(ParamSetInfo) Info
    real, intent(in):: inCls(:,:)
    logical, intent(in) :: HasCls 
	logical use_cosmonet
    real acl(lmax,num_cls)
    integer error

    GetLogLikePost  = GetLogPrior(CMB, Info)
    if ( GetLogLikePost >= logZero) then
       GetLogLikePost = logZero
       
    else 

       if (Use_HST) GetLogLikePost = GetLogLikePost + (CMB%H0 - 72)**2/(2*8**2)  !HST 
       if (Use_BBN) GetLogLikePost = GetLogLikePost + (CMB%ombh2 - 0.022)**2/(2*0.002**2) 
          !I'm using increased error bars here
   
       if (Use_CMB .or. Use_LSS) then
          if (HasCls) then
           acl = inCls
           error =0
          else
           call GetCls(CMB, Info, acl, error)
          end if
         if (error /= 0) then
          GetLogLikePost = logZero 
         else
          if (Use_CMB) GetLogLikePost = CMBLnLike(acl, use_cosmonet) + GetLogLikePost
          if (Use_mpk) GetLogLikePost = GetLogLikePost + LSSLnLike(CMB, Info%theory)
          if (Use_WeakLen) GetLogLikePost = GetLogLikePost + WeakLenLnLike(CMB, Info%theory)     
          if (Use_Lya) GetLogLikePost = GetLogLikePost +  LSS_Lyalike(CMB, Info%Theory)
          if ( GetLogLikePost >= logZero) then
            GetLogLikePost = logZero
          end if
         end if
         if (Use_SN .and. GetLogLikePost /= logZero ) then
            if (Info%Theory%SN_loglike /= 0) then
             GetLogLikePost = GetLogLikePost + Info%Theory%SN_loglike
            else
             GetLogLikePost = GetLogLikePost + SN_LnLike(CMB)
            end if
               !Assume computed only every time hard parameters change
  
  
         end if
   
       else
         if (Use_SN) GetLogLikePost = GetLogLikePost + SN_LnLike(CMB)
       end if

      if (Use_Clusters .and. GetLogLikePost /= LogZero) then
          GetLogLikePost = GetLogLikePost + &
                 (Info%Theory%Sigma_8-0.9)**2/(2*0.05**2)
          stop 'Write your cluster prior in calclike.f90 first!'
      end if

     if (GetLogLikePost /= LogZero) GetLogLikePost = GetLogLikePost/Temperature
   

    end if

  end function GetLogLikePost

end module CalcLike
