module cosmonet
use cmbtypes
  use CAMB, only : CAMB_GetResults, CAMB_GetAge, CAMBParams, CAMB_SetDefParams,Transfer_GetMatterPower, &
          AccuracyBoost,  Cl_scalar, Cl_tensor, Cl_lensed, outNone, w_lam, &
          CAMBParams_Set,Re_OpticalDepthAtZ, MT, CAMBdata, NonLinear_Pk,&
          CAMB_GetTransfers,CAMB_FreeCAMBdata,CAMB_InitCAMBdata, CAMB_TransfersToPowers, &
          initial_adiabatic,initial_vector,initial_iso_baryon,initial_iso_neutrino, initial_iso_neutrino_vel
  implicit none

	


#define NLAYERS 3

real, allocatable :: weightsTT(:), weightsTE(:), weightsEE(:)
integer, allocatable :: labelsTT(:), labelsTE(:), labelsEE(:)

!!!!!!!!!!!!!!!!!!!!!SPECTRA WEIGHTS FILES!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

character(LEN=80) :: WeightFILETT = "tt_ascii.prm"
character(LEN=80) :: WeightFILETE = "te_ascii.prm"
character(LEN=80) :: WeightFILEEE = "ee_ascii.prm"
character(LEN=80) :: NETSTR_FILETT = "tt_ascii_netstr.prm"
character(LEN=80) :: NETSTR_FILETE = "te_ascii_netstr.prm"
character(LEN=80) :: NETSTR_FILEEE = "ee_ascii_netstr.prm"

!!!!!!!!!!!!!!!!!!!!!LIKELIHOOD WEIGHTS FILES!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

character(LEN=80) :: NETSTR_LIKE_FILE = "lhood_ascii_netstr.prm"
character(LEN=80) :: PARAM_LIKE_FILE = "lhood_ascii.prm"

!!!!!!!!!!!!!!!!!!!!!INTERPOLATION LIMITS FOR EACH PARAMETER!!!!!!!!!!!!!!!!

character(LEN=80) :: NETPARAMS = "net_params.dat"

!!!IF YOU WANT ONLY COSMONET SPECTRA SET user_cosmonet = .true. AND user_cosmonet_like = .false. !!!!
!!!IF YOU WANT COSMONET LIKELIHOOD SET BOTH TO .true. !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

logical :: user_cosmonet = .true.
logical :: user_cosmonet_like = .true.

contains




subroutine init_cambnetLike(time)
!use global
use cmbtypes
integer l(1:60)
integer paramlike(1:8)                                                 
integer n
logical time

time = .false.            
                                                                                                                                                
open(43, FILE = NETSTR_LIKE_FILE)
read(43, *) paramlike
                                                                                                                                                            
netlike%nlayers = paramlike(1)
netlike%nin = paramlike(2)
netlike%nhid = paramlike(3)
netlike%nhid2 = paramlike(4)
netlike%nout = paramlike(5)
netlike%nweights = paramlike(6)
netlike%nparams = paramlike(7)
netlike%nnodes = paramlike(8)

open(44, FILE = PARAM_LIKE_FILE)
read(44,*) weightslike

write(*,*) "done with initlike"
                                                                                                                                                            
end  subroutine init_cambnetLike


 subroutine init_cambnet(netTT, netTE, netEE, weightsTT, weightsTE, weightsEE, labelsTT, labelsTE, labelsEE)
use cmbtypes
Type(NetStr)  netTT, netTE, netEE
real, allocatable :: weightsTT(:)
real, allocatable :: weightsTE(:)
real, allocatable :: weightsEE(:)
integer l(1:60)
integer, allocatable :: labelsTT(:)
integer, allocatable :: labelsTE(:)
integer, allocatable :: labelsEE(:)
integer paramTT(1:8)
integer paramTE(1:8)
integer paramEE(1:8)

integer n

!!!TT
open(13, FILE = NETSTR_FILETT)
read(13, *) paramTT

open(23, FILE = NETSTR_FILETE)
read(23, *) paramTE

open(33, FILE = NETSTR_FILEEE)
read(33, *) paramEE


netTT%nlayers = paramTT(1)
netTT%nin = paramTT(2)
netTT%nhid = paramTT(3)
netTT%nhid2 = paramTT(4)
netTT%nout = paramTT(5)
netTT%nweights = paramTT(6)
netTT%nparams = paramTT(7)
netTT%nnodes = paramTT(8)

netTE%nlayers = paramTE(1)
netTE%nin = paramTE(2)
netTE%nhid = paramTE(3)
netTE%nhid2 = paramTE(4)
netTE%nout = paramTE(5)
netTE%nweights = paramTE(6)
netTE%nparams = paramTE(7)
netTE%nnodes = paramTE(8)

netEE%nlayers = paramEE(1)
netEE%nin = paramEE(2)
netEE%nhid = paramEE(3)
netEE%nhid2 = paramEE(4)
netEE%nout = paramEE(5)
netEE%nweights = paramEE(6)
netEE%nparams = paramEE(7)
netEE%nnodes = paramEE(8)

open(14, FILE = WeightFILETT)
allocate(weightsTT(1:netTT%nparams))
read(14,*) weightsTT

open(24, FILE = WeightFILETE)
allocate(weightsTE(1:netTE%nparams))
read(24,*) weightsTE

open(34, FILE = WeightFILEEE)
allocate(weightsEE(1:netEE%nparams))
read(34,*) weightsEE


write(*,*) "INIT FINISHED"

end  subroutine init_cambnet



function regress_like()
use cmbtypes
real regress_like
integer lmax, num_Cls, i,j
real parameters(1:6)


real scalein(1:netlike%nin)
real offin(1:netlike%nin)
real scaleout(1:netlike%nout)
real offout(1:netlike%nout)
real inputs(1:netlike%nin)
real hidden1(1:netlike%nhid)
real outputs(1:netlike%nout)
real theta1(1:netlike%nhid)
real theta2(1:netlike%nout)
real w1(1:netlike%nin*netlike%nhid)
real w2(1:netlike%nhid*netlike%nout)

do i=1, num_params_used
parameters(i) = COSMONETP(i)
end do
                    

     j=1
        do i=1, netlike%nhid
                theta1(i)=weightslike(j);
                j=j+1
        end do

                                                                                 
        do i=1, netlike%nin*netlike%nhid
                w1(i)=weightslike(j);
                j=j+1
        end do


        do i=1, netlike%nout
                theta2(i)=weightslike(j);
                j=j+1
        end do


        do i=1, netlike%nhid*netlike%nout
                w2(i)=weightslike(j);
                j=j+1
        end do


!! apply feedforward calculation
!! hidden layer:
!        do i=1, netlike%nhid
!              hidden1(i)=theta1(i)
!                do j=1, netlike%nin
!                        hidden1(i) = hidden1(i) + (parameters(j)*w1(j + i*netlike%nin))
!                end do
!                hidden1(i)=tanh(hidden1(i))
!        end do
do i=1, 50
	hidden1(i) = theta1(i)
	do j=1, netlike%nin
		hidden1(i) = hidden1(i) + (parameters(j)*w1(j+(i-1)*netlike%nin))
	end do
	hidden1(i) = tanh(hidden1(i))
end do
!write(*,*) "hidden1", hidden1
outputs = 0d0
!! output layer
!        do i=1, netlike%nout
!                outputs(i)=theta2(i)
!                do j=1, netlike%nhid
!                        outputs(i) = outputs(i) + & 
!			 (hidden1(j)*w2( j+i*netlike%nhid ))
!                end do
!        end do
outputs(1) = theta2(1)
	do j=1, netlike%nhid
		outputs(1) = outputs(1) + (hidden1(j)*w2( j ))
	end do

!write(*,*) "outputs", outputs
!! scale the results
        do i=1, netlike%nout
                regress_like = outputs(i)
		!write(*,*) "regress_like", regress_like
	end do




end function regress_like


subroutine regress_cambnet(P, weights, net, Cl, lmax, num_Cls, l)
use cmbtypes
integer i,j,  num_Cls, lmax
Type(NetStr) net
type(CAMBParams)  P
real Cl(lmax,1:num_Cls)
integer l(1:60)





real parameters(1:6)
real weights(1:net%nweights)
real scalein(1:net%nin)
real offin(1:net%nin)
real scaleout(1:net%nout)
real offout(1:net%nout)
real inputs(1:net%nin)
real hidden1(1:net%nhid)
real outputs(1:net%nout)
real theta1(1:net%nhid)
real theta2(1:net%nout)
real w1(1:net%nin*net%nhid)
real w2(1:net%nhid*net%nout)



do i=1, num_params_used
parameters(i) = COSMONETP(i)
end do


      j=1
        do i=1, net%nhid
                theta1(i)=weights(j);
                j=j+1
	end do
!write(*,*) "tt_theta1", theta1

        do i=1, net%nin*net%nhid
                w1(i)=weights(j);
                j=j+1
	end do

!write(*,*) "tt_w1", w1
	do i=1, net%nout
                theta2(i)=weights(j);
                j=j+1
	end do
!write(*,*) "tt_theta2", theta2       
	do i=1, net%nhid*net%nout
                w2(i)=weights(j);
                j=j+1
        end do

!! apply feedforward calculation
!! hidden layer:
        do i=1, net%nhid
              hidden1(i)=theta1(i)
                do j=1, net%nin
                        hidden1(i) = hidden1(i) + (parameters(j)*w1(j + (i-1)*net%nin))
		end do
                hidden1(i)=tanh(hidden1(i))
        end do
!write(*,*) "tt_hidden1", hidden1
!! output layer
	do i=1, net%nout
                outputs(i)=theta2(i)
		do j=1, net%nhid
                        outputs(i) = outputs(i) + (hidden1(j)*w2(j+(i-1)*net%nhid))
		end do
	end do
!write(*,*) "tt_outputs", outputs
!! scale the results
	do i=1, net%nout
                Cl(i, num_Cls) = outputs(i)
		Cl(i,num_Cls) = Cl(i,num_Cls)/(l(i)*(l(i)+1))
		!Cl(i,num_Cls) = Cl(i,num_Cls)/((l(i)*(l(i)+1))*2*3.14159)
		!write(*,*) l(i), Cl(i,1), Cl(i,2), Cl(i,3)
	end do

!write(*,*) Cl

end subroutine 






end module cosmonet

